a {
  font-variant: initial;
}

b {
  font-variant: inherit;
}

c {
  font-variant: unset;
}

d {
  font-variant: normal;
}

e {
  font-variant: small-caps;
}

f {
  font-variant-ligatures: common-ligatures discretionary-ligatures;
  font-variant-numeric: lining-nums tabular-nums diagonal-fractions;
  font-variant-east-asian: simplified;
}
