/*
 * Copyright © 2020 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include "gdkdeviceprivate.h"

G_BEGIN_DECLS

void gdk_macos_device_query_state (GdkDevice        *device,
                                   GdkSurface        *surface,
                                   GdkSurface       **child_surface,
                                   double           *win_x,
                                   double           *win_y,
                                   GdkModifierType  *mask);

void gdk_macos_device_set_implicit_grab (GdkDevice  *device,
                                         GdkSurface *surface);

GdkSurface * gdk_macos_device_get_implicit_grab (GdkDevice *device);

G_END_DECLS
