paned separator,
picture {
  all: unset;
  background: firebrick;
}

paned.vertical > separator {
  min-width: 10px;
  min-height: 20px;
}

paned.horizontal > separator {
  min-width: 20px;
  min-height: 10px;
}
