box {
  font: 10px "Comic Sans";
}

#label1 {
  font-size: smaller;
}

#label2 {
  font: 10pt Cantarell, sans-serif;
  line-height: normal;
}

#label3 {
  font: bold italic 80% monospace;
  line-height: 1.5;
}

#label4 {
  font: expanded oblique xx-small serif;
  font-variant-caps: normal;
  font-variant-numeric: normal;
  font-variant-alternates: normal;
  font-variant-position: normal;
  font-variation-settings: normal;
  font-feature-settings: normal;
  font-kerning: normal;
  font-size: 10px;
  font-weight: bolder;
  line-height: 12px;
}

#label5 {
  font: 20mm '21st Century';
  font-variant-caps: unicase;
  font-variant-numeric: tabular-nums diagonal-fractions;
  font-variant-alternates: historical-forms;
  font-variant-position: super;
  font-variation-settings: 'wght' 300, 'slnt' 1;
  font-feature-settings: 'liga' on, 'dlig' off, 'ss01' 1, 'tnum' 0;
  font-kerning: auto;
  font-size: larger;
  line-height: 120%;
}
