.nobackground {
  background: none;
}

.shadow {
  box-shadow: 0 3px 9px 1px rgba(0, 0, 0, 0.5);
  margin: 20px;
}

.black {
  background: black;
}

.small-button {
  min-width: 16px;
  min-height: 16px;
  padding: 0;
}
