a {
  line-height: initial;
}

b {
  line-height: inherit;
}

c {
  line-height: unset;
}

d {
  line-height: normal;
}

e {
  line-height: 1.5;
}

f {
  line-height: 200%;
}

g {
  line-height: 2em;
}
