#pragma once

#include "gskcaironode.h"

G_BEGIN_DECLS

void                    gsk_cairo_node_set_surface              (GskRenderNode          *node,
                                                                 cairo_surface_t        *surface);

G_END_DECLS
