#pragma once

#include "gtkcellareabox.h"

/* Private interaction with GtkCellAreaBoxContext */
gboolean    _gtk_cell_area_box_group_visible (GtkCellAreaBox  *box,
                                              int              group_idx);

