box {
  -gtk-icon-palette: default;
  -gtk-icon-transform:  none;
}

label {
  -gtk-icon-palette: error #ffaabb, warning yellow, blue #0000bb;
  -gtk-icon-transform: scale3d(1,2,3) skew(20deg,1rad) perspective(100cm)
                       matrix3d(0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15);
}

image {
  -gtk-icon-transform: matrix(1,0, 0, 1.0, 1.5, 2.3) translate(20px) scale(1, 2.4)
                       rotate3d(1, 2, 3, 25deg);
}

picture {
  -gtk-icon-transform: translate3d(1px, 2px, 3px) scaleX(1.0) scaleY(2.0) scaleZ(3.0)
                       skewX(20deg) skewY(-10deg) translateX(1px) translateY(2px)
                       translateZ(3px);
}
