@define-color mygreen rgb(0,1,0);

a {
  color: initial;
}

b {
  color: inherit;
}

c {
  color: unset;
}

d {
  color: currentColor;
}

e {
  color: transparent;
}

f {
  color: red;
}

g {
  color: rgba(0,1,0,1);
}

h {
  color: rgb(1,0,1);
}

i {
  color: #abc;
}

j {
  color: #0123;
}

k {
  color: #012345;
}

l {
  color: #FFEEDDCC;
}

m {
  color: lighter(red);
}

n {
  color: darker(red);
}

o {
  color: shade(green,0.5);
}

p {
  color: alpha(green,0.5);
}

q {
  color: mix(red,blue,0.25);
}

r {
  color: @mygreen;
}

s {
  color: rgb(255 0 0);
}

t {
  color: rgb(255% 0 0);
}

u {
  color: rgb(255 0 0 / 50%);
}

v {
  color: rgb(calc(128 + 127) calc(10 - 10) calc(100 * 0) / calc(1 / 2));
}

w {
  color: rgba(255 0 0);
}

x {
  color: rgba(255 0 0 / 0.5);
}

y {
  color: color(srgb 1 0 0);
}

z {
  color: color(srgb 99% 98.5% 98% / 0.5);
}

aa {
  color: color(srgb 1.5 -150% 0 / 200%);
}

ab {
  color: color(srgb-linear 1 0.5 0 / 50%);
}

ac {
  color: oklab(0.8 0 0.3 / 0.5);
}

ad {
  color: oklch(0.8 0.3 200 / 0.5);
}

ae {
  color: hwb(200 10 90 / 0.5);
}

af {
  color: oklab(0% 0% 0% / 0%);
}

ag {
  color: oklab(100% 100% 100% / 100%);
}

ah {
  color: color(display-p3 0.5 0.5 0.5);
}

ai {
  color: color(xyz 0.5 0.5 0.5);
}

aj {
  color: color(rec2020 0.5 0.5 0.5);
}

ak {
  color: color(rec2100-pq 0.5 0.5 0.5);
}
