a {
  margin-left: NaN;
  margin-right: NaN;
  margin-top: NaN;
  padding-left: 3;
  padding-right: 3;
  padding-top: 2;
}

b {
  margin-left: NaN;
  margin-right: NaN;
  margin-top: -1px;
  padding-left: 3;
  padding-right: 3;
  padding-top: -3;
}

c {
  margin-bottom: 15px;
  margin-left: 20px;
  margin-right: 15px;
  margin-top: 20px;
}

d {
  margin-left: 20px;
  margin-right: 0;
  margin-top: -1px;
}

e {
  margin-bottom: NaN;
  margin-left: NaN;
  margin-right: infinite;
  margin-top: 0;
  padding-bottom: 0;
  padding-left: infinite;
  padding-right: -infinite;
  padding-top: 0;
}

f {
  margin-bottom: NaN;
  margin-left: NaN;
  margin-right: NaN;
}
